<?php
/*
Copyright (c) 2005 Luke Plant <L.Plant.98@cantab.net>

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
associated documentation files (the "Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject 
to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial
 portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN 
NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/**
 * Simple but powerful flatfile database
 *
 * The Flatfile class and related WhereClause classes can be used for 
 * manipulating tab-delimited text files.  It is designed to be a replacement
 * for a real SQL database when the performance of a real database
 * is not needed, or when in the prototyping stage.
 * It supports a very powerful and extensible equivalent
 * to the WHERE clause.  It provides some equivalent
 * to SELECT ... ORDER BY (ordering by a single field only),
 * and no implementation of JOINs or GROUP BY.  
 *
 * @tutorial flatfile.pkg
 * @todo File locking
 * @package flatfile
 * @license http://www.opensource.org/licenses/mit-license.php
 */
 
  /** Used to indicate a comparison should be done using PHPs builtin method for comparisons. */
define('AUTO_COMPARISON', 0);
 /** Used to indicate a comparison should be done as a string comparison (vars will be converted to strings before comparing). */
define('STRING_COMPARISON', 1);
 /** Used to indicate a comparison should be done as an integer comparison (vars will be converted to integers before comparing).  */
define('INTEGER_COMPARISON', 2);

/** Indicates ascending order */
define('ASCENDING', 1);
/** Indicates descending order */
define('DESCENDING', -1);


/**
 * Implements simple but powerful flat file database storage and retrieval
 *
 * Includes equivalents to SELECT * FROM table WHERE..., DELETE WHERE ...
 * UPDATE and more.  All files are stored in the {@link Flatfile::$datadir $datadir} directory,
 * and table names are just filenames in that directory.  Subdirectories
 * can be used just by specifying a table name that includes the directory name.
 * @package flatfile
 */
class Flatfile {
	/** @access private */
	var $tables;
	
	/** The directory to store files in.  */
	var $datadir;

	/**
	 * Get all rows from a table
	 * @param string $tablename	The table to get rows from
	 * @return array The table as an array of rows, where each row is an array of columns
	 */
	function selectAll ($tablename) {
		if (!isset($this->tables[$tablename]))
			$this->loadTable($tablename);
		return $this->tables[$tablename];	
	}
	
	/**
	 * Selects rows from a table that match the specified criteria
	 *
	 * This simulates the following SQL query:
	 * <pre>
	 *   SELECT TOP $limit * FROM  $tablename
	 *   WHERE $whereclause
	 *   ORDER BY $orderByField $orderByType
	 * </pre>
	 *
	 * @param string $tablename	The table (file) to get the data from
	 * @param object $whereclause 	Either a {@link WhereClause WhereClause} object to do selection of rows, or NULL to select all
	 * @param int $limit  		The max number of rows to return or -1 for all
	 * @param int $orderByField	The index (in the table row array) of the field to sort by (or -1 to omit the ORDER BY clause)
	 * @param int $orderByType	Either ASCENDING or DESCENDING (ignored if $orderByField == -1)
	 * @return array The matching data, as an array of rows, where each row is an array of columns
	 */
	function selectWhere ($tablename, $whereclause, $limit = -1, $orderByField = -1, $orderByType = ASCENDING) {
		if (!isset($this->tables[$tablename]))
			$this->loadTable($tablename);
		
		$table = $this->selectAll($tablename); // Get a copy
		
		if ($orderByField != -1) {
			// Create a function to use for sorting.
			// If we had closures, and usort supported them,
			// this would be much nicer
			$n = $orderByField;
			$function = create_function('$a,$b',
			'if ($a[' . $n . '] == $b['.$n.']) { 
			 	return 0; 
			} else {
				
			 	return ($a['.$n.'] < $b['.$n.']) ? '. -$orderByType .' :  '. $orderByType .' ;
			}	');
			usort($table, $function);
		}
		
		$results = array();
		$count = 0;

		foreach ($table as $row) {
			if ($whereclause === NULL || $whereclause->testRow($row)) {
				$results[] = $row;
				if ($limit != -1 && (++$count >= $limit))
					break;
			}
		}
		return $results;
	}
	
	/**
	 * Select a row using a unique ID
	 * @param string $tablename 	The table to get data from
	 * @param string $idField	The index of the field containing the ID
	 * @param string $id		The ID to search for
	 * @return array 	The row of the table as an array
	 */
	function selectUnique ($tablename, $idField, $id) {
		if (!isset($this->tables[$tablename]))
			$this->loadTable($tablename);
		foreach ($this->tables[$tablename] as $row) {
			if ($row[$idField] == $id) {	
				return $row;
			}
		}
		return array();
	}
	
	/**
	 * Inserts a row with an automatically generated ID
	 *
	 * The autogenerated ID will be the highest ID in the column so far plus one.
	 * @param string $tablename	The table to insert data into
	 * @param int $idField		The index of the field which is the ID field
	 * @param array $newRow		The new row to add to the table
	 * @return int		The newly assigned ID	 
	 */
	function insertWithAutoId ($tablename, $idField, $newRow) 
	{
		$rows = $this->selectWhere($tablename, null, 1, $idField, -1);
		if ($rows) {
			$newId = $rows[0][$idField] + 1;
		} else {
			$newId = 1;
		}
		$newRow[$idField] = $newId;
		$this->tables[$tablename][] = $newRow;
		$this->writeTable($tablename);
		return $newId;
	}

	
	/**
	 * Inserts a row in a table
	 *
	 * @param string $tablename	The table to insert data into
	 * @param array $newRow		The new row to add to the table
	 */
	function insert ($tablename, $newRow) 
	{
		if (!isset($this->tables[$tablename]))
			$this->loadTable($tablename);
			
		$this->tables[$tablename][] = $newRow;
		$this->writeTable($tablename);
	}
	
	/**
	 * Updates an existing row using a unique ID
	 *
	 * @param string $tablename	The table to update
	 * @param int $idField		The index of the field which is the ID field
	 * @param array $updatedRow	The updated row to add to the table
	 */
	function updateRowById ($tablename, $idField, $updatedRow)
	{
		if (!isset($this->tables[$tablename]))
			$this->loadTable($tablename);
		for ($i = 0; $i < count($this->tables[$tablename]); ++$i) {
			if ($this->tables[$tablename][$i][$idField] == $updatedRow[$idField]) {
				$this->tables[$tablename][$i] = $updatedRow;
			}
		}
		$this->writeTable($tablename);
	}
	
	/**
	 * Deletes all rows in a table that match specified criteria
	 *
	 * @param string $tablename	The table to update
	 * @param object $whereClause.  {@link WhereClause WhereClause} object that will select rows to be
	 *   deleted.  All rows are deleted if $whereClause == NULL
	 */
	function deleteWhere ($tablename, $whereClause) {
		if (!isset($this->tables[$tablename]))
			$this->loadTable($tablename);
		
		for ($i = count($this->tables[$tablename]) - 1; $i >= 0 ; --$i) {
			if ($whereClause === NULL || $whereClause->testRow($this->tables[$tablename][$i])) {
				unset($this->tables[$tablename][$i]);
			}
		}
		$this->writeTable($tablename);		
		$this->loadTable($tablename); // reset indexes
	}
	
	/**
	 * Delete all rows in a table
	 *
	 * @param string $tablename	The table to update
	 */
	function deleteAll ($tablename) {
		$this->deleteWhere($tablename, NULL);
	}

	/**#@+
	 * @access private
	 */
	function loadTable ($tablename) {
		$filedata = file($this->datadir . $tablename);
		$table = array();
		if (is_array($filedata)) {
			foreach ($filedata as $line) {
				$line = rtrim($line, "\n");
				$table[] = explode("\t", $line);
			}
		}
		$this->tables[$tablename] = $table;
	}

	function writeTable ($tablename) {
		$output = '';
		
		foreach ($this->tables[$tablename] as $row) {
			for ($i = 0; $i < count($row); ++$i) {
				if ($i > 0) $output .= "\t";
				$output .= str_replace(array("\t","\r","\n"), array(''), $row[$i]);
			}
			$output .= "\n";
		}
		$fp = @fopen($this->datadir . $tablename, "w");
		fwrite($fp, $output, strlen($output));
		fclose($fp);
	}
	/**#@-*/
}

/////////////////////////// WHERE CLAUSE CLASSES ////////////////////////////////////

/**
 * Used to test rows in a database table, like the WHERE clause in an SQL statement.
 *
 * @abstract
 * @package flatfile
 */
class WhereClause 
{
	/**
	 * Tests a table row object
	 *
	 * @param array $row   The row to test
	 * @return bool True if the $row passes the WhereClause
	 * selection criteria, false otherwise
	 */
	function testRow ($row) {}
}

/**
 * Implements a single WHERE clause that does simple comparisons of a field
 * with a value.
 *
 * @package flatfile
 */
class SimpleWhereClause extends WhereClause 
{
	/**#@+
	 * @access private
	 */
	var $field;
	var $operator;
	var $value;
	var $compare_type;
	/**#@-*/

	/**
	 * Creates a new {@link WhereClause WhereClause} object that does a comparison
	 * of a field and a value.
	 *
	 * This will be the most commonly used type of WHERE clause.  It can do comparisons
	 * of the sort "$tablerow[$field] operator $value"
	 * where 'operator' is one of:<br>
	 * - = (equals)</li>
	 * - != (not equals)</li>
	 * - > (greater than)</li>
	 * - < (less than)</li>
	 * 
	 * @param int $field        	The index (in the table row) of the field to test
	 * @param string $operator  	The comparison operator, one of "=", "!=", "<", ">"
	 * @param mixed $value		The value to compare to.
	 * @param int $compare_type	The comparison method to use - either
	 *  AUTO_COMPARISON (default), STRING_COMPARISON or INTEGER_COMPARISON
	 *
	 */
	function SimpleWhereClause ($field, $operator, $value, $compare_type = STRING_COMPARISON)
	{
		$this->field = $field;
		$this->operator = $operator;
		$this->value = $value;
		$this->compare_type = $compare_type;
	}
	
	function testRow ($tablerow) {
		if ($this->field < 0)
			return TRUE;
		$compval = $this->value;
		$fieldval = $tablerow[$this->field];
		
		if ($this->compare_type == STRING_COMPARISON) {
			$compval = (string)$compval;
			$fieldval = (string)$fieldval;
		} else if ($this->compare_type == INTEGER_COMPARISON) {
			$compval = (int)$compval;
			$fieldval = (int)$fieldval;
		}
		
		if (($this->operator == '=')  && ($fieldval == $compval))
			return TRUE;
		if (($this->operator == '!=') && ($fieldval != $compval))
			return TRUE;
		if (($this->operator == '>')  && ($fieldval > $compval))
			return TRUE;
		if (($this->operator == '<')  && ($fieldval < $compval))
			return TRUE;
		return FALSE;	
	}
}

/**
 * {@link WhereClause WhereClause} class to match a value from a list of items
 * @package flatfile
 */
class ListWhereClause extends WhereClause {

	/** @access private */
	var $field;
	/** @access private */
	var $list;

	
	/**
	 * Creates a new ListWhereClause object
	 *
	 * The resulting WhereClause will pass rows (return true) if the value of the specified
	 * field is in the array.
	 * @todo allow XXX_COMPARISON to be passed, and implement
	 * 
	 * @param int $field		Field to match
	 * @param array $list		List of items
	 */
	function ListWhereClause ($field, $list) {
		$this->list = $list;
		$this->field = (int)$field;
	}
	
	function testRow ($tablerow) {
		return in_array($tablerow[$this->field], $this->list);
	}
}

/**
 * Abstract class that combines zero or more {@link WhereClause WhereClause} objects
 * together.
 * @package flatfile
 */ 
class CompositeWhereClause extends WhereClause
{
	/** @access private */
	var $clauses = array();
	
	/**
	 * Add a {@link WhereClause WhereClause} to the list of clauses to be used for testing
	 *
	 * @param object $whereClause	The WhereClause object to add
	 */ 
	function add ($whereClause)
	{
		$this->clauses[] = $whereClause;
	}
}

/**
 * {@link CompositeWhereClause CompositeWhereClause} that does an OR on all its
 * child WhereClauses.
 *
 * Use the {@link CompositeWhereClause::add() add()} method to add WhereClause objects
 * to the list of clauses to check.  The testRow function of the resulting object
 * will then return true if any of its child clauses return true (and returns
 * false if no clauses have been added for consistency with that statement).
 * @package flatfile
 */
class OrWhereClause extends CompositeWhereClause 
{
	function testRow ($tablerow) {
		foreach ($this->clauses as $clause) {
			if ($clause->testRow($tablerow))
				return true;
		}
		return false;
	}
}

/**
 * {@link CompositeWhereClause CompositeWhereClause} that does an AND on all its
 * child WhereClauses.
 *
 * Use the {@link CompositeWhereClause::add() add()} method to add WhereClause objects
 * to the list of clauses to check.  The testRow function of the resulting object
 * will then return false if any of its child clauses return false (and returns
 * true if no clauses have been added for consistency with that statement).
 * @package flatfile
 */
class AndWhereClause extends CompositeWhereClause 
{
	function testRow ($tablerow) {
		foreach ($this->clauses as $clause) {
			if (!$clause->testRow($tablerow))
				return false;
		}
		return true;
	}
}


